/*
 * 	game.c
 * 	Veci tykajici se vlastni hry
 *
 * 	Josef Kufner <jkufner@seznam.cz>
 */
#define _REENTRANT
#define _XOPEN_SOURCE

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <pthread.h>
#include <semaphore.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <time.h>

#include "game.h"
#include "interface2.h"
#include "netgate.h"

#define DEBUG 		10
#include "debug.h"

#define	PLAY_AREA(r, c)		(play_area[(r) * game_cfg.col + (c)])
#define PLAYER_CFG(p)		((p) == PLAYER_X ? game_cfg.x : game_cfg.o )
#define PLAYER2_CFG(p)		((p) == PLAYER_X ? game_cfg.o : game_cfg.x )
#define TURN_FORMAT		"> row = %d, col = %d;"

/* Datovy typ pro historii tahu */
typedef struct turn_history_struct {
	int player;		/* Kdo tahnul */
	int turn;		/* Kam tahnul */

	struct turn_history_struct *next, *prev;
} TURN_HISTORY;

/* Prvni polozka seznamu historie tahu */
TURN_HISTORY *turn_hist_first = NULL;

/* Aktualni tah pri prochazeni hry */
static TURN_HISTORY *game_pos = NULL;

/* Hraci plocha (dynamicky alokovane jednorozm. pole) */
int *play_area = NULL;

/* Tah predavany hraci deskou */
struct {
	int now;
	int turn;
} pa_turn = { 0, 0 };

/* Probiha hra ? */
int is_game_now = 0;

/* Rizeni prubehu hry */
static int action = 0;
static int player = PLAYER_X;

/* Konfigurace hry */
GAME_CFG game_cfg = {
	{ 1, "Kek", "/usr/local/bin/piskv-rbt -b -s 25 -n x -a 1", 1, 
		"127.0.0.1", 5544, 1, 
		0,
		"/tmp/piskv-pipe1", NULL, NULL, -1 }, 
	{ 1, "Koleko", "/usr/local/bin/piskv-rbt -s 25 -n o -a 1", 0, 
		"127.0.0.1", 5544, 0, 
		0,
		"/tmp/piskv-pipe2", NULL, NULL, -1 },
	PLAYER_NONE,
	"127.0.0.1", 5544, "piskv-netgate",
	25, 25, 1,
	PLAYER_X, 0
};

/* Pomocne pole pro urcovani kontrolu na vitezstvi */
 const int smer[4][2] = {
	 /*	row,	col	*/
	 { 	0, 	1 	},	/* Vpravo */
	 { 	1, 	1 	},	/* Sikmo vpravo */
	 { 	1, 	0 	},	/* Dolu */
	 { 	1, 	-1 	},	/* Sikmo vlevo */
 };

/* Lokalni pomocne funkce */
static int Check(int player, int row, int col, int row_inc, int col_inc);
static int Recv_Turn (int player);
static int Send_Turn(int player, int turn);
static int Create_Play_Array (void);


/*****************************************************************************
 * 	Start Game  -  Zahajeni hry
 */
void Start_Game (void) {
 char str[1024];
	
 	DBG_MSG(2, "Start_Game...");

	Create_Play_Array();
	
	Turn_Hist_Clear();
	Hist_List_Update();

	/* Otevirani rour, spoustreni robotu, navazovani sitovych spojeni,...*/
	for (player = PLAYER_X;; player = PLAYER_O) {

		sprintf(PLAYER_CFG(player).rbt_f_nam, "/tmp/piskv-rbt-%c-%ld", 
				PLAYER_SYMBOL(player), (long int) getpid());
	
		/* Mistni robot */
		if (PLAYER_CFG(player).is_local && PLAYER_CFG(player).is_rbt) {
			/* Tvorba roury */
			DBG_MSG(3, "Spousteni robotu (hrac %c) ...", PLAYER_SYMBOL(player));
			remove(PLAYER_CFG(player).rbt_f_nam);
			mkfifo(PLAYER_CFG(player).rbt_f_nam, 0777);
			
			/* Pousteni robota s presmerovanym vstupem */
			sprintf(str, "%s > %s", PLAYER_CFG(player).cmd, PLAYER_CFG(player).rbt_f_nam);
			DBG_MSG(6, "Hrac %c, cmd = \"%s\"", PLAYER_SYMBOL(player), str);
			PLAYER_CFG(player).rbt_f_out = popen(str, "w");
			DBG_MSG(6, "Hrac %c, rbt_f_out = %p", PLAYER_SYMBOL(player), PLAYER_CFG(player).rbt_f_out);
			if (PLAYER_CFG(player).rbt_f_out == NULL) {
				perror("Nemohu spustit robota");
				exit(1);
			}

			/* Otevreni vystupu robota */
			PLAYER_CFG(player).rbt_f_in = fopen(PLAYER_CFG(player).rbt_f_nam, "r");
			DBG_MSG(6, "Hrac %c, rbt_f_in = %p", PLAYER_SYMBOL(player), PLAYER_CFG(player).rbt_f_in);
			if (PLAYER_CFG(player).rbt_f_in == NULL) {
				perror("Nemohu otevrit rouru");
				exit(1);
			}
		}

		/* Vzdaleny hrac */
		if (! PLAYER_CFG(player).is_local) {
			/* Uprava prikazoveho radku */
			if (PLAYER_CFG(player).is_server) {
				/* Server */
				sprintf(str, "%s %s %s %s %s %s %ld > %s", 
						game_cfg.netgate_cmd,
						NETGATE_P_SERVER,
						(PLAYER_CFG(player).adr[0] == '*') ?
							NETGATE_P_ACCEPT_ALL : "",
						NETGATE_P_ADDRESS,
						PLAYER_CFG(player).adr,
						NETGATE_P_PORT,
						PLAYER_CFG(player).port,
						PLAYER_CFG(player).rbt_f_nam);
			} else {
				/* Client */
				sprintf(str, "%s %s %s %s %s %ld > %s", 
						game_cfg.netgate_cmd,
						NETGATE_P_CLIENT,
						NETGATE_P_ADDRESS,
						PLAYER_CFG(player).adr,
						NETGATE_P_PORT,
						PLAYER_CFG(player).port,
						PLAYER_CFG(player).rbt_f_nam);
			}
			
			/* Tvorba roury */
			DBG_MSG(3, "Spousteni netgate (hrac %c) ...", PLAYER_SYMBOL(player));
			remove(PLAYER_CFG(player).rbt_f_nam);
			mkfifo(PLAYER_CFG(player).rbt_f_nam, 0777);
			
			/* Pousteni netgate s presmerovanym vstupem */
			DBG_MSG(6, "Hrac %c, cmd = \"%s\"", PLAYER_SYMBOL(player), str);
			PLAYER_CFG(player).rbt_f_out = popen(str, "w");
			DBG_MSG(6, "Hrac %c, rbt_f_out = %p", PLAYER_SYMBOL(player), PLAYER_CFG(player).rbt_f_out);
			if (PLAYER_CFG(player).rbt_f_out == NULL) {
				perror("Nemohu spustit netgate");
				exit(1);
			}

			/* Otevreni vystupu netgate */
			PLAYER_CFG(player).rbt_f_in = fopen(PLAYER_CFG(player).rbt_f_nam, "r");
			DBG_MSG(6, "Hrac %c, rbt_f_in = %p", PLAYER_SYMBOL(player), PLAYER_CFG(player).rbt_f_in);
			if (PLAYER_CFG(player).rbt_f_in == NULL) {
				perror("Nemohu otevrit rouru");
				exit(1);
			}
		}
		
		if (player == PLAYER_O)
			break;
	}
	
	action = 0;
	player = game_cfg.begin;
	game_pos = 0;
	game_cfg.winner = PLAYER_NONE;
	game_pos = NULL;
}


/*****************************************************************************
 * 	End_Game  -  Konec Hry
 */
void End_Game (void) {
 TURN_HISTORY *cur;
	
	DBG_MSG(3, "End_Game...");
	
	Hist_List_Update();
	
	/* Likvidace rour a ukoncovani robotu */
	for (player = PLAYER_X;; player = PLAYER_O) {
		if ((PLAYER_CFG(player).is_local && PLAYER_CFG(player).is_rbt)
				|| (!PLAYER_CFG(player).is_local)) {
			/* Prikaz k ukonceni */
			putc(0, PLAYER_CFG(player).rbt_f_out);
			fclose(PLAYER_CFG(player).rbt_f_in);
			remove(PLAYER_CFG(player).rbt_f_nam);
			pclose(PLAYER_CFG(player).rbt_f_out);
		}
		
		if (player == PLAYER_O)
			break;
	}

	for (cur = turn_hist_first; cur != NULL; cur = cur->next)
		Hist_List_Move(HL_NEXT);
	
	DBG_MSG(3, "End Game - Done.")	
	/* Konec hry */
	is_game_now = 0;
	refresh_btns();
}


/*****************************************************************************
 * 	Turn  -  Provedeni tahu, rizeni prubehu hry
 *
 * 	urceno pro opakovane volani
 */
int Turn (void) {
 char str[512];
 static int turn = 0, last_turn = -1;
	if (! is_game_now) {
		return -1;
	}
 
 	switch (action) {
		case 0:		/* Zobrazeni zpravy o cekani na hrace */
			DBG_MSG(3, "");
			DBG_MSG(3, "Turn - ekm na hre %s (%c)......", 
					PLAYER_CFG(player).name, PLAYER_SYMBOL(player));
			if (last_turn != -1) {
				sprintf(str, "Posledn tah byl \"row = %d, col = %d\". ekm na hre %s (%c)...", 
						last_turn / game_cfg.col, last_turn % game_cfg.row,
						PLAYER_CFG(player).name, PLAYER_SYMBOL(player));
			} else {
				sprintf(str, "ekm na hre %s (%c)...", 
						PLAYER_CFG(player).name, PLAYER_SYMBOL(player));
			}
			Message(str);
			pa_turn.now = 0;
			action++;
			/* break chybi umyslne */
			
		case 1:		/* Prijem tahu */
			if ((turn = Recv_Turn(player)) >= 0) {
				pa_turn.now = 0;
				DBG_MSG(3, "Turn - Prijaty tah: turn = %d", turn);
				last_turn = turn;
				action++;
			}
			break;
			
		case 2:		/* Poslani tahu */
			DBG_MSG(3, "Turn - Poslani tahu...");
			if (player == PLAYER_X)
				Send_Turn(PLAYER_O, turn);
			else
				Send_Turn(PLAYER_X, turn);
			DBG_MSG(5, "Turn - Poslany tah: turn = %d", turn);
			action++;
			break;

		case 3:		/* Pridani tahu do historie */
			DBG_MSG(3, "Turn - Pridani tahu do historie...");
			Turn_Hist_Add_Item(player, turn);
			
			/* Tady lze provedst refresh historie - zdrzuje */
			/* Hist_List_Update(); */
			
			action++;
			/* break chybi umyslne */
			
		case 4:		/* Kontrola tahu */
			DBG_MSG(3, "Turn - Kontrola tahu...");
			if (play_area[turn] == PLAYER_NONE) {
				play_area[turn] = player;
				action++;
			} else {
				DBG_MSG(3, "Turn - Tah na obsazene pole !");
				End_Game();
				
				sprintf(str, "Tah hre %s (%c) byl na obsazen pole !", 
						PLAYER_CFG(player).name, 
						PLAYER_SYMBOL(player));
				Message(str);
				action = 0;
			}
			break;
			
		case 5:		/* Zobrazeni tahu */
			DBG_MSG(3, "Zobrazeni tahu...");
			Show_Turn(turn, player);
			action++;
			break;
			
		case 6:		/* Kontrola konce hry se zobrazenim*/
			DBG_MSG(3, "Kontrola konce hry...");
			
			if (Check_Play_Area(player))
				End_Game();
			
			/* Prepnuti hrace */
			player = (player == PLAYER_X) ? PLAYER_O : PLAYER_X;
			action++;

			action = 0;
			break;
			
		default:
			fprintf(stderr, "Fatal: Interni chyba programu - funkce Turn");
			exit(1);
	}
	return 0;
}


/*****************************************************************************
 * 	Check  -  Kontrola vitezstvi se zobrazenim
 */
int Check(int player, int row, int col, int row_inc, int col_inc) {
 int i;
 char str[512];

	/* Kontrola vyskytu petice */
	for (i = 0; i < 5; i++) {
		if (PLAY_AREA(row + i * row_inc, col + i * col_inc) != player
				|| row + i * row_inc >= game_cfg.row 
				|| col + i * col_inc >= game_cfg.col
				|| row + i * row_inc < 0 
				|| col + i * col_inc < 0) {
			return 0;
		}
	}

	for (i = 0; i < 5; i++) {
		Highlight_Turn(player, (row + i * row_inc) * game_cfg.col 
				+ (col + i * col_inc)); 
	}

	sprintf(str, "Hr %s (%c) vyhrl !", 
			PLAYER_CFG(player).name, PLAYER_SYMBOL(player));
	game_cfg.winner = player;
	Message(str);
	return 1;
}


/*****************************************************************************
 * 	Check_Play_Area  -  Kontrola na vitezstvi funkci Check
 */
int Check_Play_Area (int player) {
 int r, c, s;
	
	for (r = 0; r < game_cfg.row; r++) {
		for (c = 0; c < game_cfg.col; c++) {
			for (s = 0; s < 4; s++) {
				if (Check(player, r, c, smer[s][0], smer[s][1])) {
					return 1;
				}
			}
		}
	}
	return 0;
}


/*****************************************************************************
 * 	Recv_Turn  -  Prijem tahu
 */
int Recv_Turn (int player) {
 int r, c;

	/* Mistni hrac  */
	if (PLAYER_CFG(player).is_local && ! PLAYER_CFG(player).is_rbt) {
		if (pa_turn.now)
			return pa_turn.turn;
		else
			return -1;
	}

	/* Mistni robot nebo netgate */
	if ((PLAYER_CFG(player).is_local && PLAYER_CFG(player).is_rbt) 
			|| (!PLAYER_CFG(player).is_local)) {
		fscanf(PLAYER_CFG(player).rbt_f_in, "> row = %d, col = %d;", &r, &c);
		DBG_MSG(5, "Recv_Turn: Prijato: row = %d, col = %d -> turn = %d", 
				r, c, r * game_cfg.col + c);
		getc(PLAYER_CFG(player).rbt_f_in);
		return r * game_cfg.col + c;
	}

	return -1;	
}


/*****************************************************************************
 * 	Send_Turn  -  Poslani Tahu
 */
int Send_Turn(int player, int turn) {
 int r, c;
	
	/* Mistni hrac  -  Nic se nedeje */
	if (PLAYER_CFG(player).is_local && ! PLAYER_CFG(player).is_rbt) {
		return 0;
	}	
	
	/* Mistni robot nebo netgate */
	if ((PLAYER_CFG(player).is_local && PLAYER_CFG(player).is_rbt)
			|| (!PLAYER_CFG(player).is_local)) {
		/* Prepocet souradnic */
		r = turn / game_cfg.col;
		c = turn % game_cfg.row;
		
		DBG_MSG(5, "Send_Turn: player = %c, row = %d, col = %d", 
				PLAYER_SYMBOL(player), r, c);
		fprintf(PLAYER_CFG(player).rbt_f_out, "> row = %d, col = %d;\n", 
				r, c);
		fflush(PLAYER_CFG(player).rbt_f_out);
	}
	
	return 0;
}


/*****************************************************************************
 * 	Create_Play_Area  -  Vytvoreni nove hraci plochy
 */
int Create_Play_Array (void) {
 int i;

	/* Uvolneni pameti, pokud je alokovana */
	if (play_area)
		free(play_area);

	/* Alokace nove pameti dane velikosti */
	play_area = (int*) malloc (game_cfg.row * game_cfg.row * sizeof(int));
	if (play_area == NULL) {
		perror("malloc failed");
		exit(1);
	}

	/* Nastaveni poli na prazdne */
	for (i = 0; i < game_cfg.row * game_cfg.col; i++)
		play_area[i] = PLAYER_NONE;

	return 0;
}

/*****************************************************************************
 * 	Play_Area_Turn  -  Predani tahu od desky
 */
void Play_Area_Turn (int turn) {
	pa_turn.now = 1;
	pa_turn.turn = turn;
}


/*****************************************************************************
 * 	Turn_Hist_Add_Item  -  Pridani tahu do seznamu
 */
int Turn_Hist_Add_Item (int player, int turn) {
 static TURN_HISTORY *cur = NULL;
 static TURN_HISTORY *prev= NULL;

	/* Prvni polozka seznamu */ 	
 	if (turn_hist_first == NULL) {
		/* Jsme na zacatku -  tvorba 1. polozky */
		/* Alokace pameti */
		turn_hist_first = (TURN_HISTORY*) malloc(sizeof(TURN_HISTORY));
		if (turn_hist_first == NULL) {
			perror("malloc failed");
			return -1;
		}

		/* Nastaveni hodnot polozky */
		turn_hist_first->prev = NULL;
		turn_hist_first->next = NULL;
		turn_hist_first->player = player;
		turn_hist_first->turn = turn;

		/* Nastaveni polohy v seznamu */
		prev = turn_hist_first;
		game_pos = turn_hist_first;
		cur = NULL;
	} else {
		/* Tvorba dalsi polozky */
		/* Alokace pameti*/
		cur = (TURN_HISTORY *) malloc(sizeof(TURN_HISTORY));
		if (cur == NULL) {
			perror("malloc failed");
			return -1;
		}

		/* Nastaveni hodnot */
		prev->next = cur;
		cur->prev = prev;
		cur->next = NULL;
		cur->player = player;
		cur->turn = turn;

		/* Nastaveni pozice v seznamu */
		prev = cur;
		game_pos = cur;
		cur = NULL;
	}
	/* Nastaveni pozice pro prehravani */
	return 0;
}

/******************************************************************************
 * 	Turn_Hist_Get_Item  -  Precteni jedne polozky seznamu
 */
int Turn_Hist_Get_Item (const int where, int *player, int *turn) {
 static TURN_HISTORY *cur = NULL;
	
	switch (where) {
		case TH_START:
			/* Nastaveni pozice v seznamu na zacatek */
			cur = turn_hist_first;			
		case TH_NEXT:
			/* Kontrola konce seznamu */
			if (cur == NULL)
				return 1;
 
			/* Cteni hodnot hodnot */
			*player = cur->player;
			*turn = cur->turn;

			/* Posun v seznamu */
			cur = cur->next;

			break;
		default:
			return -1;
	}
 
	return 0;
}


/******************************************************************************
 * 	Turn_Hist_Clear  -  Vymazani seznamu
 */
void Turn_Hist_Clear(void) {
 TURN_HISTORY *cur, *next = NULL;

	/* Likvidace polozky */
	for (cur = turn_hist_first; cur != NULL; cur = next) {
		next = cur->next;
		free(cur);
	}

	/* Nastaveni konce seznamu */
	turn_hist_first = NULL;
	return;
}


/******************************************************************************
 *  	Load_Game  -  Nacteni prubehu hry ze souboru
 */
int Load_Game (const char *file_name) {
 FILE *file;
 char str[512], line[512], ch;
 int val, r, c, i, t, p;
 int clear_msg = 1;

	Message("Natm soubor...");
 
 	/* Vymazani soucasne hry */
	Turn_Hist_Clear();
	Destroy_Play_Area();
	
	/* Otevreni souboru */
	file = fopen(file_name, "r");
	if (file == NULL) {
		perror("Load_Game: Open failed");
		return -1;
	}
	
	/* Naplneni seznamu historie tahu */
	while (! feof(file)) {
		ch = getc(file);
		if (feof(file))
			break;

		/* Vynechani prazdnych radku */
		if (ch == '\n')
			continue;
		
		/* Vynechani komentaru */
		if (ch == '#') {
			while (!feof(file) && getc(file) != '\n')
				;
			continue;
		}

		/* Nacteni radky */
		line[0] = ch;			/* Vlozeni prvniho znaku */
		for (i = 1; !feof(file) && (ch = getc(file)) != '\n' && i < 511; i++)
			line[i] = ch;
		line[i] = '\0';			/* Ukonceni retezce */

		DBG_MSG(7, "Load line: \"%s\"", line)
			
		/* Cteni Konfigurace (radek uvozeny tabulatorem) */
		if (line[0] == '\t') {
			/* dekodovani informaci */
			if (sscanf(line + 1, "%s = %d", str, &val) != 2) {
				/* Error */
				Message("Chyba v souboru !");
				clear_msg = 0;
				break;
			}
			DBG_MSG(6, "Load: cfg: %s = %d", str, val);
			if (strcmp(str, "rows") == 0) {
				game_cfg.row = val;
			} else if (strcmp(str, "cols") == 0) {
				game_cfg.col = val;
			}

			continue;
		}

		/* Cteni tahu */
		if (line[0] == 'X' || line[0] == 'O') {
			/* Dekodovani tahu */
			if (sscanf(line + 1, ":\t%d,\t%d", &r, &c) != 2) {
				/* Error */
				Message("Chyba v souboru !");
				clear_msg = 0;
				break;
			}
			val = r * game_cfg.col + c;
			DBG_MSG(6, "Load: turn: %c: row = %d, col = %d -> turn = %d", line[0], r, c, val);

			/* Tvorba tahu */
			Turn_Hist_Add_Item(line[0] == PLAYER_SYMBOL(PLAYER_X) ? PLAYER_X : PLAYER_O, val);
		}
	}

	/* Zavreni souboru */
	if(fclose(file) == EOF) {
		perror("Save_Game: fclose failed");
		return -1;
	}

	/* Refresh GUI */
	if (clear_msg)
		Message("");
	Hist_List_Update();
	Create_Play_Area();		/* Tvorba mista v pameti */
	Create_Play_Array();		/* Tvorba GUI */
	
	/* Zobrazeni nactene hry */
	r = Turn_Hist_Get_Item(TH_START, &p, &t);
	while (r == 0) {
		DBG_MSG(7, "Load: Zobrazuji tah %d(%c)...", t, PLAYER_SYMBOL(p));
		play_area[t] = p;
		Show_Turn(t, p);
		r = Turn_Hist_Get_Item(TH_NEXT, &p, &t);
	}
	Check_Play_Area(PLAYER_X);
	Check_Play_Area(PLAYER_O);
	return 0;
}


/******************************************************************************
 * 	Save_Game  -  Ulozeni prubehu hry do souboru
 */
int Save_Game (const char *file_name) {
 FILE *file;
 TURN_HISTORY *cur;
 char t_str[256];
 time_t the_time;
 struct tm *tm_ptr;

	/* Otevreni souboru */
	file = fopen(file_name, "w");
	if (file == NULL) {
		/* Chyba */
		perror("Save_Game: Open failed");
		return -1;
	}
	
	/* Zjisteni aktualniho casu */
	(void) time(&the_time);
	tm_ptr = localtime(&the_time);
	strftime(t_str, 60, "%A %d.%m.%Y %H:%M:%S", tm_ptr);
		
	/* Zapis dat */
	fprintf(file,	"# Hra pikvorky - zznam hry\n"
			"#\n"
			"# Uloeno v %s\n"
			"# Hri:\n"
			"#  X: %s (%s)\n"
			"#  O: %s (%s)\n"
			"# Vtz:%s%c\n"
			"#\n\n", 
			t_str,
			game_cfg.x.name, game_cfg.x.is_local ? 
				(game_cfg.x.is_rbt ? "robot" : "mstn")
				: "vzdlen",
			game_cfg.o.name, game_cfg.o.is_local ? 
				(game_cfg.o.is_rbt ? "robot" : "mstn")
				: "vzdlen",
			game_cfg.winner == PLAYER_NONE ? "dn" : " ",
			PLAYER_SYMBOL(game_cfg.winner)); 

	fprintf(file, "# Velikost hrac plochy:\n"
			"\trows = %d\n\tcols = %d\n\n", 
			game_cfg.row, game_cfg.col);
	
	fprintf(file, "# Prbh hry:\n"
			"#\trow,\tcol\n");
	
	for (cur = turn_hist_first; cur != NULL; cur = cur->next) {
		fprintf(file, "%c:\t%d,\t%d\n", 
				PLAYER_SYMBOL(cur->player), 
				cur->turn / game_cfg.col,
				cur->turn % game_cfg.row);
	}
	fprintf(file, "\n# End.\n");
	
	/* Zavreni souboru */
	if(fclose(file) == EOF) {
		perror("Save_Game: fclose failed");
		return -1;
	}
	return 0;
}


/******************************************************************************
 * 	Next_Turn  -  Dalsi tah
 *
 * 	Vraci 1 pokud existuje dalsi polozka nebo 0 pokud jsme na konci.
 */
int Next_Turn (void) {
 int i, win;
 char str[512];
 
	/* Pokud neni game_pos nastaveno, tak ji nastavime */
 	if (game_pos == NULL) {
		game_pos = turn_hist_first;
	}

	/* Posun pozice */
	if (game_pos->next != NULL) {

		game_pos = game_pos->next;
		Hist_List_Move(HL_NEXT);

		Remove_Highlight();
		play_area[game_pos->turn] = game_pos->player;
		for (i = 0; i < game_cfg.row * game_cfg.col; i++) {
			Show_Turn(i, play_area[i]);
		}
		win = Check_Play_Area(game_pos->player);
	} else {
		return 0;
	}
	sprintf(str, "Tah \"row = %d, col = %d\", hr %s(%c)%s", 
			game_pos->turn / game_cfg.col, game_pos -> turn % game_cfg.row, 
			PLAYER_CFG(game_pos->player).name, PLAYER_SYMBOL(game_pos->player),
			win ? " vyhrl." : ".");
	Message(str);
	return 1;
}


/******************************************************************************
 * 	Prev_Turn  -  Predchozi tah
 *
 * 	Vraci 1 pokud existuje predchozi polozka nebo 0 pokud jsme na zacatku.
 */
int Prev_Turn (void) {
 int i, win;
 char str[512];
	
	/* Pokud neni game_pos nastaveno, tak ji nastavime */
 	if (game_pos == NULL) {
		game_pos = turn_hist_first;
	}

	/* Posun pozice */
	if (game_pos->prev != NULL) {
		/* Predchudce existuje */
		Remove_Highlight();
		play_area[game_pos->turn] = PLAYER_NONE;
		for (i = 0; i < game_cfg.row * game_cfg.col; i++) {
			Show_Turn(i, play_area[i]);
		}
		
		win = Check_Play_Area(game_pos->player);
		
		game_pos = game_pos->prev;
		Hist_List_Move(HL_PREV);
	} else {
		return 0;
	}
	sprintf(str, "Tah \"row = %d, col = %d\", hr %s(%c)%s", 
			game_pos->turn / game_cfg.col, game_pos -> turn % game_cfg.row, 
			PLAYER_CFG(game_pos->player).name, PLAYER_SYMBOL(game_pos->player),
			win ? " vyhrl." : ".");
	Message(str);
	return 1;
}


